#!/bin/bash
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=1
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=2
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=3
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=4
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=5
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=6
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=7
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=8
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=9
python -m imitation.scripts.train_imitation bc with bc_seals_ant seed=10
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=1
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=2
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=3
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=4
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=5
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=6
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=7
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=8
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=9
python -m imitation.scripts.train_imitation dagger with dagger_seals_ant seed=10
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=1
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=2
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=3
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=4
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=5
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=6
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=7
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=8
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=9
python -m imitation.scripts.train_adversarial airl with airl_seals_ant seed=10
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=1
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=2
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=3
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=4
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=5
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=6
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=7
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=8
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=9
python -m imitation.scripts.train_adversarial gail with gail_seals_ant seed=10
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=1
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=2
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=3
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=4
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=5
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=6
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=7
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=8
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=9
python -m imitation.scripts.train_imitation bc with bc_seals_half_cheetah seed=10
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=1
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=2
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=3
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=4
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=5
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=6
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=7
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=8
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=9
python -m imitation.scripts.train_imitation dagger with dagger_seals_half_cheetah seed=10
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=1
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=2
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=3
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=4
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=5
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=6
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=7
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=8
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=9
python -m imitation.scripts.train_adversarial airl with airl_seals_half_cheetah seed=10
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=1
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=2
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=3
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=4
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=5
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=6
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=7
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=8
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=9
python -m imitation.scripts.train_adversarial gail with gail_seals_half_cheetah seed=10
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=1
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=2
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=3
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=4
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=5
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=6
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=7
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=8
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=9
python -m imitation.scripts.train_imitation bc with bc_seals_hopper seed=10
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=1
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=2
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=3
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=4
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=5
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=6
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=7
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=8
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=9
python -m imitation.scripts.train_imitation dagger with dagger_seals_hopper seed=10
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=1
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=2
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=3
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=4
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=5
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=6
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=7
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=8
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=9
python -m imitation.scripts.train_adversarial airl with airl_seals_hopper seed=10
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=1
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=2
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=3
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=4
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=5
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=6
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=7
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=8
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=9
python -m imitation.scripts.train_adversarial gail with gail_seals_hopper seed=10
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=1
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=2
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=3
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=4
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=5
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=6
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=7
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=8
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=9
python -m imitation.scripts.train_imitation bc with bc_seals_swimmer seed=10
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=1
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=2
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=3
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=4
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=5
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=6
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=7
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=8
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=9
python -m imitation.scripts.train_imitation dagger with dagger_seals_swimmer seed=10
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=1
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=2
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=3
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=4
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=5
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=6
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=7
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=8
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=9
python -m imitation.scripts.train_adversarial airl with airl_seals_swimmer seed=10
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=1
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=2
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=3
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=4
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=5
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=6
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=7
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=8
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=9
python -m imitation.scripts.train_adversarial gail with gail_seals_swimmer seed=10
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=1
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=2
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=3
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=4
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=5
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=6
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=7
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=8
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=9
python -m imitation.scripts.train_imitation bc with bc_seals_walker seed=10
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=1
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=2
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=3
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=4
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=5
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=6
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=7
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=8
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=9
python -m imitation.scripts.train_imitation dagger with dagger_seals_walker seed=10
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=1
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=2
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=3
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=4
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=5
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=6
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=7
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=8
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=9
python -m imitation.scripts.train_adversarial airl with airl_seals_walker seed=10
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=1
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=2
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=3
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=4
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=5
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=6
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=7
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=8
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=9
python -m imitation.scripts.train_adversarial gail with gail_seals_walker seed=10
